#include "gcoder.h"
#include <limits>
#include <QGLWidget>
#include "mesh.h"
#include <QDebug>
#include <QTextStream>
#include "printsettings.h"
#include <iostream>
#include <QProgressDialog>
#include <QEventLoop>
#include <QTime>
#include <QMutexLocker>
#include "streamwriter.h"

GCoder *GCoder::pInstance = NULL;

GCoder *GCoder::instance()
{
    if (!pInstance)
        pInstance = new GCoder;
    return pInstance;
}

GCoder::GCoder() : mutex(QMutex::Recursive)
{
}

void GCoder::clear()
{
	lock();
    layers.clear();
    cached_lines.clear();
    cached_points.clear();
	cached_color.clear();
	unlock();
}

void GCoder::draw()
{
	lock();
	if (cached_lines.empty())
    {
		cached_color.clear();
#pragma omp parallel
		{
			std::vector<Vec3f> local_cached_lines;
			std::vector<Vec3f> local_cached_color;
#pragma omp for nowait
			for(auto layer = layers.begin() ; layer < layers.end() ; ++layer)
			{
				const double z = layer->first;
				Vec2d prev(std::numeric_limits<double>::infinity(),
						   std::numeric_limits<double>::infinity());
				size_t i = local_cached_lines.size();
				for(const Vec2d &p : layer->second)
				{
					if (p.x() == std::numeric_limits<double>::infinity()
							|| prev.x() == std::numeric_limits<double>::infinity())
					{
						prev = p;
						continue;
					}
					local_cached_lines.push_back(Vec3f(prev.x(), prev.y(), z));
					local_cached_lines.push_back(Vec3f(p.x(), p.y(), z));
					prev = p;
				}
				const float n = local_cached_lines.size() - i;
				for(float k = 0.f ; i < local_cached_lines.size() ; ++i, ++k)
					local_cached_color.push_back(Vec3f(k/n, k/n, k/n));
			}
#pragma omp critical
			{
				cached_lines.insert(cached_lines.end(), local_cached_lines.begin(), local_cached_lines.end());
				cached_color.insert(cached_color.end(), local_cached_color.begin(), local_cached_color.end());
			}
		}
    }
	glEnable(GL_VERTEX_ARRAY);
	glEnable(GL_COLOR_ARRAY);
	glDisable(GL_NORMAL_ARRAY);
	glDisable(GL_TEXTURE_COORD_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, cached_lines.data());
	glColorPointer(3, GL_FLOAT, 0, cached_color.data());
	glDrawArrays(GL_LINES, 0, cached_lines.size());

	glDisable(GL_COLOR_ARRAY);
    glColor3ub(0,0,0xFF);
    glEnable(GL_POINT_SMOOTH);
    glPointSize(4.0);
    if (cached_points.empty())
    {
		size_t n = 0;
		for(const auto &layer : layers)
			n += layer.second.size();
		cached_points.reserve(n);
		for(const auto &layer : layers)
        {
            const double z = layer.first;
            for(const Vec2d &p : layer.second)
            {
                if (p.x() == std::numeric_limits<double>::infinity())
                    continue;
                cached_points.push_back(Vec3f(p.x(), p.y(), z));
            }
        }
    }
    glVertexPointer(3, GL_FLOAT, 0, cached_points.data());
    glDrawArrays(GL_POINTS, 0, cached_points.size());
    glDisable(GL_POINT_SMOOTH);
    glPointSize(1.0);

	glDisable(GL_VERTEX_ARRAY);
	unlock();
}

void GCoder::computeLayers(Mesh *mesh)
{
	QMutexLocker mLock(&mutex);
    const double layer_height = std::max(PrintSettings::instance()->getLayerHeight(), 1e-16);
    clear();

    if (!mesh)
        return;

    std::pair<Vec3, Vec3> bbox = mesh->getBoundingBox();
    const double z_max = PrintSettings::instance()->getZmax();
    const double z_max_i = z_max - int(z_max / layer_height) * layer_height > 0.01 * layer_height ? z_max + layer_height : z_max;
    bbox.second.z() = std::min(z_max_i, bbox.second.z());
	for(double z = 0.0 ; z <= bbox.second.z() ; z += layer_height)
        layers.push_back(std::make_pair(z, std::vector<Vec2d>()));

	mesh->getBoundingBox();		// Cache the result

	mesh->buildKdTree();
	mesh->computeFaceNormals();
	mesh->buildSupportHull();

	QProgressDialog *pDialog = new QProgressDialog(QObject::tr("Generating slices"), QString(), 0, layers.size());
	pDialog->show();

	volatile size_t next = 0;
#pragma omp parallel
	while(next < layers.size())
    {
		size_t i;
#pragma omp critical
		{
			i = next;
			++next;
		}
		if (i >= layers.size())
			break;
#pragma omp master
		{
			pDialog->setValue(i);
			pDialog->repaint();
		}
		mesh->slice(layers[i].first, layer_height, layers[i].second, i);
    }
    Vec2d last(0,0);
    for(auto &l : layers)
        last = Mesh::optimizeSlice(last, l.second);

	delete pDialog;
}

QByteArray GCoder::computeGCodeFromLayers() const
{
	QMutexLocker mLock(&mutex);
    const int extrusion_speed_nominal = int(PrintSettings::instance()->getExtrusionSpeed());
    const int extrusion_speed_at_start = int(PrintSettings::instance()->getExtrusionSpeedAtStart());
    const double retraction_on_move = PrintSettings::instance()->getRetractionOnMove();
    const double support_density_ratio = PrintSettings::instance()->getSupportDensityRatio();

	QTime timer;
	timer.start();

	StreamWriter gcode(0x1000000);

    gcode << "G21\n";
    gcode << "G90\n";
    gcode << "G92 X0 Y0 Z0 E0\n";
    gcode << "G28\n";
    gcode << "M83\n";
    gcode << "G4 P0\n";
	gcode << "G1 Z10 F4000\n";
    gcode << "G1 E30 F" << extrusion_speed_at_start << '\n';
    gcode << "G4 P1000\n";
    gcode << "G1 Z0 F4000\n";

    const double zoffset = PrintSettings::instance()->getZOffset();
    const double extrusion_multiplier = PrintSettings::instance()->getExtrusionMultiplier();
    const double extruded_width = PrintSettings::instance()->getExtrudedWidth();
    const bool b_z_move = PrintSettings::instance()->isZmoveEnabled();
    const double x_scale = PrintSettings::instance()->getXscale();
    const double y_scale = PrintSettings::instance()->getYscale();
    const double z_scale = PrintSettings::instance()->getZscale();

	Vec2d prev(0,0);
    for(size_t i = 0 ; i < layers.size() ; ++i)
    {
        const int extrusion_speed = i < 3 ? extrusion_speed_at_start : extrusion_speed_nominal;
        const int free_move_speed = std::max(10000, extrusion_speed);
        const int retraction_speed = 10000;
        int speed = retraction_speed;
        gcode << "G1 E-0.4 F" << retraction_speed << '\n';
        gcode << "G1 X" << prev.x() * x_scale << " Y" << prev.y() * y_scale << " Z" << std::max<double>(layers[i].first, zoffset) * z_scale << '\n';
        gcode << "G4 P0\n";
        bool b_new_line = true;
        bool b_support_line = false;
        bool b_z_moved = false;
        for(const auto &it : layers[i].second)
            if (it.x() != std::numeric_limits<double>::infinity())
            {
				const Vec2d p = it;
                if (std::round(prev.x() * 1e2) == std::round(p.x() * 1e2)
                        && std::round(prev.y() * 1e2) == std::round(p.y() * 1e2))
                {
                    b_new_line = false;
                    continue;
                }
                if (b_new_line)
                {
                    if (b_z_move && (prev - p).length() > extruded_width * 2.0)
                    {
                        gcode << "G1 Z" << std::max<double>(layers[i].first + 1.0, zoffset) * z_scale;
                        if (speed != free_move_speed)
                            gcode << " F" << free_move_speed;
                        speed = free_move_speed;
                        gcode << '\n';
                        b_z_moved = true;
                    }
                }
                gcode << "G1 X" << p.x() * x_scale << " Y" << p.y() * y_scale;
                if (!b_new_line)
                {
                    gcode << " E" << (prev - p).length() * extrusion_multiplier * (b_support_line ? support_density_ratio : 1.0);
                    if (speed != extrusion_speed)
                        gcode << " F" << extrusion_speed;
                    speed = extrusion_speed;
                }
                else
                {
                    if (speed != free_move_speed)
                        gcode << " F" << free_move_speed;
                    speed = free_move_speed;
                }
                gcode << '\n';
                if (b_new_line)
                {
                    if (b_z_move && b_z_moved)
                        gcode << "G1 Z" << std::max<double>(layers[i].first, zoffset) * z_scale << '\n';
                    b_z_moved = false;
                    if (retraction_on_move > 0.0)
                    {
                        gcode << "G1 E" << retraction_on_move << " F" << retraction_speed << '\n';
                        speed = retraction_speed;
                    }
                }
				b_new_line = false;
                prev = p;
            }
            else
            {
                b_support_line = (it.y() == -std::numeric_limits<double>::infinity());
                if (!b_new_line)
                {
                    b_new_line = true;
                    if (retraction_on_move > 0.0)
                    {
                        gcode << "G1 E-" << retraction_on_move;
                        if (speed != retraction_speed)
                            gcode << " F" << retraction_speed;
                        speed = retraction_speed;
                        gcode << '\n';
                    }
                }
            }
    }

	qDebug() << timer.elapsed() << "ms";

	return QByteArray(gcode.data(), gcode.size());
}

namespace
{
	void parseCoord(const QString &line, Vec3 &pos, double &E, const double coef, double *F = NULL)
	{
		const QStringList &chunks = line.split(' ');
		for(int i = 1 ; i < chunks.size() ; ++i)
		{
			const QString &s = chunks[i];
            switch(s[0].toLatin1())
			{
			case 'X':
			case 'x':
				pos.x() = coef * (s.right(s.size() - 1).toDouble());
				break;
			case 'Y':
			case 'y':
				pos.y() = coef * (s.right(s.size() - 1).toDouble());
				break;
			case 'Z':
			case 'z':
				pos.z() = coef * (s.right(s.size() - 1).toDouble());
				break;
			case 'E':
			case 'e':
				E = coef * (s.right(s.size() - 1).toDouble());
				break;
			case 'F':
			case 'f':
				if (F)
					*F = s.right(s.size() - 1).toDouble();
				break;
			}
		}
	}
}

std::pair<Vec3, Vec3> GCoder::parseGCode(const QByteArray &gcode)
{
	QMutexLocker mLock(&mutex);

	clear();

	QTextStream stream(gcode);
	stream.setLocale(QLocale("C"));

	Vec3 origin(0,0,0);
	Vec3 current_pos(0,0,0);
	bool b_absolute = true;
	double coef = 1.0;

	while(!stream.atEnd())
	{
		QString line = stream.readLine().trimmed();
		if (line.isEmpty() || line.at(0) == ';')
			continue;
		if (line.contains(';'))
			line = line.remove(line.indexOf(';'), line.size()).trimmed();

        switch(line[0].toLatin1())
		{
		case 'G':
			if (line.startsWith("G0 ") || line == "G0")
			{
				Vec3 next_pos(b_absolute ? current_pos - origin : Vec3(0,0,0));
				double E;
				parseCoord(line, next_pos, E, coef);
				if (b_absolute)
					current_pos = next_pos + origin;
				else
					current_pos += next_pos;
			}
			else if (line.startsWith("G1 ") || line == "G1")
			{
				Vec3 next_pos(b_absolute ? current_pos - origin : Vec3(0,0,0));
				double E(-1.0);
				parseCoord(line, next_pos, E, coef);
				if (E > 0.0)
				{
					cached_lines.push_back(Vec3f(current_pos.x(), current_pos.y(), current_pos.z()));
					cached_points.push_back(Vec3f(current_pos.x(), current_pos.y(), current_pos.z()));
				}
				if (b_absolute)
					current_pos = next_pos + origin;
				else
					current_pos += next_pos;
				if (E > 0.0)
				{
					cached_lines.push_back(Vec3f(current_pos.x(), current_pos.y(), current_pos.z()));
					cached_points.push_back(Vec3f(current_pos.x(), current_pos.y(), current_pos.z()));
				}
			}
			else if (line.startsWith("G28 ") || line == "G28")
			{
				Vec3 next_pos(std::numeric_limits<double>::infinity(),
							  std::numeric_limits<double>::infinity(),
							  std::numeric_limits<double>::infinity());
				double E;
				parseCoord(line, next_pos, E, coef);
				if (next_pos[0] == std::numeric_limits<double>::infinity()
						&& next_pos[1] == std::numeric_limits<double>::infinity()
						&& next_pos[2] == std::numeric_limits<double>::infinity())
				{
					current_pos = origin = Vec3(0,0,0);
				}
				else
					for(size_t i = 0 ; i < 3 ; ++i)
						if (next_pos[i] != std::numeric_limits<double>::infinity())
						{
							current_pos[i] = 0;
							origin[i] = 0;
						}
			}
			else if (line.startsWith("G20 ") || line == "G20")
			{
				coef = 25.4;
			}
			else if (line.startsWith("G21 ") || line == "G21")
			{
				coef = 1.0;
			}
			else if (line.startsWith("G90 ") || line == "G90")
			{
				b_absolute = true;
			}
			else if (line.startsWith("G91 ") || line == "G91")
			{
				b_absolute = false;
			}
			else if (line.startsWith("G92 ") || line == "G92")
			{
				Vec3 next_pos(current_pos);
				double E;
				parseCoord(line, next_pos, E, coef);
				origin = current_pos - next_pos;
			}
			break;
		case 'M':
			if (line.startsWith("M206 ") || line == "M206")
			{
				double E;
				parseCoord(line, origin, E, coef);
			}
			break;
		default:
			break;
		}
	}

	for(size_t i = 0 ; i < cached_lines.size() ; ++i)
		cached_color.push_back(Vec3f(1,1,1) * (float(i) / cached_lines.size()));

	std::pair<Vec3, Vec3> bbox(Vec3(std::numeric_limits<double>::infinity(),
									std::numeric_limits<double>::infinity(),
									std::numeric_limits<double>::infinity()),
							   -Vec3(std::numeric_limits<double>::infinity(),
									 std::numeric_limits<double>::infinity(),
									 std::numeric_limits<double>::infinity()));
	for(const auto &p : cached_points)
	{
		for(size_t i = 0 ; i < 3 ; ++i)
		{
			bbox.first[i] = std::min<double>(bbox.first[i], p[i]);
			bbox.second[i] = std::max<double>(bbox.second[i], p[i]);
		}
	}
	return bbox;
}

double GCoder::estimateTimeToPrint(const QByteArray &gcode) const
{
	QTextStream stream(gcode);
	stream.setLocale(QLocale("C"));

	Vec3 origin(0,0,0);
	Vec3 current_pos(0,0,0);
	bool b_absolute = true;
	double coef = 1.0;
	double F = 0.0;
	double t = 0.0;

	while(!stream.atEnd())
	{
		QString line = stream.readLine().trimmed();
		if (line.isEmpty() || line.at(0) == ';')
			continue;
		if (line.contains(';'))
			line = line.remove(line.indexOf(';'), line.size()).trimmed();

        switch(line[0].toLatin1())
		{
		case 'G':
			if (line.startsWith("G0 ") || line == "G0")
			{
				Vec3 next_pos(b_absolute ? current_pos - origin : Vec3(0,0,0));
				double E;
				parseCoord(line, next_pos, E, coef, &F);
				const Vec3 start_pos = current_pos;
				if (b_absolute)
					current_pos = next_pos + origin;
				else
					current_pos += next_pos;
				t += (current_pos - start_pos).length() / F * 60.0;
			}
			else if (line.startsWith("G1 ") || line == "G1")
			{
				Vec3 next_pos(b_absolute ? current_pos - origin : Vec3(0,0,0));
				double E(-1.0);
				parseCoord(line, next_pos, E, coef, &F);
				const Vec3 start_pos = current_pos;
				if (b_absolute)
					current_pos = next_pos + origin;
				else
					current_pos += next_pos;
				t += (current_pos - start_pos).length() / F * 60.0;
			}
			else if (line.startsWith("G28 ") || line == "G28")
			{
				Vec3 next_pos(std::numeric_limits<double>::infinity(),
							  std::numeric_limits<double>::infinity(),
							  std::numeric_limits<double>::infinity());
				double E;
				parseCoord(line, next_pos, E, coef);
				if (next_pos[0] == std::numeric_limits<double>::infinity()
						&& next_pos[1] == std::numeric_limits<double>::infinity()
						&& next_pos[2] == std::numeric_limits<double>::infinity())
				{
					current_pos = origin = Vec3(0,0,0);
				}
				else
					for(size_t i = 0 ; i < 3 ; ++i)
						if (next_pos[i] != std::numeric_limits<double>::infinity())
						{
							current_pos[i] = 0;
							origin[i] = 0;
						}
			}
			else if (line.startsWith("G20 ") || line == "G20")
			{
				coef = 25.4;
			}
			else if (line.startsWith("G21 ") || line == "G21")
			{
				coef = 1.0;
			}
			else if (line.startsWith("G90 ") || line == "G90")
			{
				b_absolute = true;
			}
			else if (line.startsWith("G91 ") || line == "G91")
			{
				b_absolute = false;
			}
			else if (line.startsWith("G92 ") || line == "G92")
			{
				Vec3 next_pos(current_pos);
				double E;
				parseCoord(line, next_pos, E, coef);
				origin = current_pos - next_pos;
			}
			break;
		case 'M':
			if (line.startsWith("M206 ") || line == "M206")
			{
				double E;
				parseCoord(line, origin, E, coef);
			}
			break;
		default:
			break;
		}
	}

	return t;
}

void GCoder::lock() const
{
	mutex.lock();
}

void GCoder::unlock() const
{
	mutex.unlock();
}
