#include "kdtree.h"
#include <algorithm>

KdTree::KdTree()
    : childs(NULL)
{
}

KdTree::KdTree(const Vec3 *vertex, Triangle *begin, Triangle *end)
    : childs(NULL)
{
    build(vertex, begin, end);
}

KdTree::~KdTree()
{
    if (childs)
        delete [] childs;
}

namespace std
{
    inline Vec3 min(const Vec3 &a, const Vec3 &b)
    {
        return Vec3(min(a.x(), b.x()),
                    min(a.y(), b.y()),
                    min(a.z(), b.z()));
    }

    inline Vec3 max(const Vec3 &a, const Vec3 &b)
    {
        return Vec3(max(a.x(), b.x()),
                    max(a.y(), b.y()),
                    max(a.z(), b.z()));
    }
}

void KdTree::build(const Vec3 *vertex, Triangle *begin, Triangle *end)
{
    this->begin = begin;
    this->end = end;
    this->vertex = vertex;

    center = Vec3(0,0,0);
    radius = 0.0;
    Vec3 lower, upper;
    if (begin != end)
    {
        upper = lower = vertex[begin->a];
        for(Triangle *t = begin ; t < end ; ++t)
        {
            lower = std::min(lower, vertex[t->a]);
            lower = std::min(lower, vertex[t->b]);
            lower = std::min(lower, vertex[t->c]);
            upper = std::max(upper, vertex[t->a]);
            upper = std::max(upper, vertex[t->b]);
            upper = std::max(upper, vertex[t->c]);
        }
        center = 0.5 * (upper + lower);
        for(Triangle *t = begin ; t < end ; ++t)
        {
            radius = std::max(radius, (center - vertex[t->a]).sq());
            radius = std::max(radius, (center - vertex[t->b]).sq());
            radius = std::max(radius, (center - vertex[t->c]).sq());
        }
        radius = std::sqrt(radius);
    }

    childs = NULL;

    if (end - begin < 4)
        return;

    upper -= lower;
    n = 0;
    if (upper[1] > upper[n])
        n = 1;
    if (upper[2] > upper[n])
        n = 2;

    auto mid0 = std::partition(begin, end, [&](const Triangle &t)
    {
        return vertex[t.a][n] < center[n]
                && vertex[t.b][n] < center[n]
                && vertex[t.c][n] < center[n];
    } );

    auto mid1 = std::partition(mid0, end, [&](const Triangle &t)
    {
        return !(vertex[t.a][n] > center[n]
                && vertex[t.b][n] > center[n]
                && vertex[t.c][n] > center[n]);
    } );

    if (mid0 == end || mid1 == begin || (mid0 == begin && mid1 == end))
        return;

    childs = new KdTree[3];
	if (end - begin > 10000)
	{
#pragma omp task
		childs[0].build(vertex, begin, mid0);
#pragma omp task
		childs[1].build(vertex, mid0, mid1);
		childs[2].build(vertex, mid1, end);
	}
	else
	{
		childs[0].build(vertex, begin, mid0);
		childs[1].build(vertex, mid0, mid1);
		childs[2].build(vertex, mid1, end);
	}
}

inline bool KdTree::rayTriangleCollision(const Vec3 &pos, const Vec3 &dir, const Vec3 &a, const Vec3 &b, const Vec3 &c, Vec3 &p) const
{
	const Vec3 n = (b - a) ^ (c - a);
    const double n_dir = n | dir;
    if (n_dir == 0.0)
        return false;
    const double f = (a - pos) | n;
    if ((f < 0.0 && n_dir > 0.0)
            || (f > 0.0 && n_dir < 0.0))
        return false;
    const Vec3 &pa = a - pos;
    const Vec3 &pb = b - pos;
    const Vec3 &pc = c - pos;
    double qa = (pb ^ (c - b)) | dir;
    double qb = (pc ^ (a - c)) | dir;
    double qc = (pa ^ (b - a)) | dir;
    if ((qa < 0.0 && qb > 0.0)
            || (qa > 0.0 && qb < 0.0))
        return false;
    if ((qb < 0.0 && qc > 0.0)
            || (qb > 0.0 && qc < 0.0))
        return false;
    if ((qc < 0.0 && qa > 0.0)
            || (qc > 0.0 && qa < 0.0))
        return false;
    const double qd = qa + qb + qc;
	p = (qa * a + qb * b + qc * c) / qd;
    return true;
}

const Triangle *KdTree::queryRay(const Vec3 &pos,
                                 const Vec3 &dir,
                                 Vec3 &hit_point,
                                 const std::unordered_set<const Triangle *> &avoid,
                                 double max_dist) const
{
    const double d2 = ((center - pos) ^ dir).sq();
    if (d2 > radius * radius)
        return NULL;
    const double oh = (center - pos) | dir;
    const double op2 = radius * radius - d2;
    if (oh < 0.0 && oh * oh > op2)
        return NULL;
    if (oh - std::sqrt(op2) > max_dist)
        return NULL;
    if (!childs)
    {
        const Triangle *t_hit = NULL;
        Vec3 p;
        for(const Triangle *t = begin ; t < end ; ++t)
        {
            if (avoid.count(t))
                continue;
            if (rayTriangleCollision(pos, dir, vertex[t->a], vertex[t->b], vertex[t->c], p))
            {
                const double l = (p - pos) | dir;
                if (l < max_dist)
                {
                    t_hit = t;
                    max_dist = l;
                    hit_point = p;
                }
            }
        }
        return t_hit;
    }

    const double d = pos[n] - center[n];
    if (d >= 0.0)
    {
        const Triangle *t_hit = childs[2].queryRay(pos, dir, hit_point, avoid, max_dist);
        if (t_hit)
            max_dist = (hit_point - pos) | dir;
        const Triangle *t2;
        if ((t2 = childs[1].queryRay(pos, dir, hit_point, avoid, max_dist)))
        {
            t_hit = t2;
            max_dist = (hit_point - pos) | dir;
        }
		if (dir[n] < 0.0)
        {
            t2 = childs[0].queryRay(pos, dir, hit_point, avoid, max_dist);
            if (t2)
                t_hit = t2;
        }
        return t_hit;
    }
    else
    {
        const Triangle *t_hit = childs[0].queryRay(pos, dir, hit_point, avoid, max_dist);
        if (t_hit)
            max_dist = (hit_point - pos) | dir;
        const Triangle *t2;
        if ((t2 = childs[1].queryRay(pos, dir, hit_point, avoid, max_dist)))
        {
            t_hit = t2;
            max_dist = (hit_point - pos) | dir;
        }
		if (dir[n] > 0.0)
        {
            t2 = childs[2].queryRay(pos, dir, hit_point, avoid, max_dist);
            if (t2)
                t_hit = t2;
        }
        return t_hit;
    }
}
