#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "serialconsole.h"
#include <QDebug>
#include <QMessageBox>
#include <QFileDialog>
#include "viewer.h"
#include "mesh.h"
#include <QSettings>
#include "gcoder.h"
#include "gcodeviewer.h"
#include "printsettings.h"

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    ui->mainToolBar->addActions(ui->menu_File->actions());
    ui->mainToolBar->addSeparator();
    ui->mainToolBar->addActions(ui->menu_Print->actions());

    connect(ui->action_Exit,SIGNAL(triggered()),this,SLOT(quit()));
    connect(ui->action_Open_mesh,SIGNAL(triggered()),this,SLOT(openMesh()));
    connect(ui->action_About,SIGNAL(triggered()),this,SLOT(about()));
    connect(ui->action_About_Qt,SIGNAL(triggered()),this,SLOT(aboutQt()));
    connect(ui->action_Stop_printing,SIGNAL(triggered()),SerialConsole::instance(),SLOT(stop()));
    connect(ui->action_Emergency_stop,SIGNAL(triggered()),SerialConsole::instance(),SLOT(emergencyStop()));
    connect(ui->action_Start_printing,SIGNAL(triggered()),this,SLOT(startPrinting()));
    connect(ui->action_Pause_printing,SIGNAL(triggered()),SerialConsole::instance(),SLOT(pause()));
    connect(ui->action_Resume_printing,SIGNAL(triggered()),SerialConsole::instance(),SLOT(unpause()));
    connect(ui->action_Next_step,SIGNAL(triggered()),SerialConsole::instance(),SLOT(step()));
	connect(ui->action_Estimate_print_time,SIGNAL(triggered()),this,SLOT(estimatePrintDuration()));

    connect(ui->bGCode,SIGNAL(pressed()),this,SLOT(updateGCode()));

    addDockWidget(Qt::RightDockWidgetArea, SerialConsole::instance());
    addDockWidget(Qt::RightDockWidgetArea, GCodeViewer::instance());
    addDockWidget(Qt::RightDockWidgetArea, PrintSettings::instance());
    tabifyDockWidget(SerialConsole::instance(),
                     GCodeViewer::instance());
    tabifyDockWidget(SerialConsole::instance(),
                     PrintSettings::instance());

    ui->view->addWidget(Viewer::instance());

    connect(SerialConsole::instance(), SIGNAL(updateWorkProgress(int)), this, SLOT(updateWorkProgress(int)));
    connect(SerialConsole::instance(), SIGNAL(updateETA(double)), this, SLOT(updateETA(double)));

	show();

	QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry", saveGeometry()).toByteArray());
	restoreState(settings.value("state", saveState()).toByteArray());
	settings.endGroup();
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::openMesh()
{
    QSettings settings;
    settings.beginGroup("Misc");
    QString dir = settings.value("last_dir", QString()).toString();
    settings.endGroup();
    const QString &filename = QFileDialog::getOpenFileName(this, tr("Open mesh"), dir, tr("Meshes (*.off *.obj *.3ds *.stl *.ply)"));
    if (filename.isEmpty())
        return;

    dir = QFileInfo(filename).absolutePath();
    settings.beginGroup("Misc");
    settings.setValue("last_dir", dir);
    settings.endGroup();

	openMesh(filename);
}

void MainWindow::openMesh(const QString &filename)
{
	Mesh *mesh = new Mesh;
    mesh->load(filename);

    GCoder::instance()->clear();
    GCodeViewer::instance()->setGCode(QByteArray());

    Viewer::instance()->setMesh(mesh);
}

void MainWindow::quit()
{
    QMessageBox::StandardButton answer = QMessageBox::warning(this, tr("Exit"), tr("Are you sure?"), QMessageBox::Yes | QMessageBox::No, QMessageBox::No);

    if (answer != QMessageBox::Yes)
        return;
	aboutToClose();
    qApp->quit();
}

void MainWindow::about()
{
    QMessageBox::about(this,tr("About QRepRap"),tr("ABOUT_MSG").arg(qApp->applicationName(), qApp->applicationVersion()));
}

void MainWindow::aboutQt()
{
    QMessageBox::aboutQt(this);
}

void MainWindow::closeEvent(QCloseEvent *e)
{
	aboutToClose();
	QMainWindow::closeEvent(e);
}

void MainWindow::aboutToClose()
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
	settings.setValue("state", saveState());
	settings.endGroup();
}

void MainWindow::updateGCode()
{
    GCoder::instance()->computeLayers(Viewer::instance()->getMesh());
	const QByteArray &gcode = GCoder::instance()->computeGCodeFromLayers();
    GCodeViewer::instance()->setGCode(gcode);
    Viewer::instance()->update();
}

void MainWindow::startPrinting()
{
    const QByteArray &gcode_data = GCodeViewer::instance()->getGCode();
    const QList<QByteArray> &lines = gcode_data.split('\n');

    foreach(const QByteArray &line, lines)
    {
        if (line.isEmpty())
            continue;
        SerialConsole::instance()->sendMessage(line);
    }
    SerialConsole::instance()->sendMessage("G1 E-20 F4000");
    SerialConsole::instance()->sendMessage("M0");
    SerialConsole::instance()->sendMessage("M0");
}

void MainWindow::updateWorkProgress(int percent)
{
    ui->pbWorkProgress->setValue(percent);
}

void MainWindow::updateETA(double s)
{
    if (s == 0.0)
        ui->statusBar->showMessage(tr("done"));
    else
    {
        ui->statusBar->showMessage(QTime(0,0).addSecs(s).toString() + tr(" remaining"));
    }
}

void MainWindow::estimatePrintDuration()
{
    const quint64 t = GCoder::instance()->estimateTimeToPrint(GCodeViewer::instance()->getGCode());
    QMessageBox::information(this,
                             tr("Estimation of print duration"),
                             tr("Print time is estimated to ")
                             + QString::number(t / 86400) + "d "
                             + QString::number((t / 3600) % 24) + "h"
                             + QString::number((t / 60) % 60) + "m"
                             + QString::number(t % 60) + "s");
}
