#include "printsettings.h"
#include "ui_printsettings.h"
#include <QSettings>
#include <QGraphicsScene>
#include <QTime>
#include <QTextStream>
#include <QGraphicsTextItem>
#include "serialconsole.h"
#include "utils.h"
#include "viewer.h"
#include "mesh.h"
#include "temperaturegraph.h"

PrintSettings *PrintSettings::pInstance = NULL;
PrintSettings *PrintSettings::instance()
{
    if (!pInstance)
        pInstance = new PrintSettings;
    return pInstance;
}

PrintSettings::PrintSettings(QWidget *parent) :
    QDockWidget(parent),
    ui(new Ui::PrintSettings)
{
    ui->setupUi(this);

    loadSettings();

    updateBuildRange();

    TemperatureGraph *T_graph = new TemperatureGraph;
    ui->temperature_frame->setLayout(new QVBoxLayout);
    ui->temperature_frame->layout()->addWidget(T_graph);
    T_graph->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);

    connect(ui->layer_height,SIGNAL(editingFinished()),this,SLOT(saveSettings()));
    connect(ui->extruder_temperature,SIGNAL(editingFinished()),this,SLOT(saveSettings()));
    connect(ui->bed_temperature,SIGNAL(editingFinished()),this,SLOT(saveSettings()));
    connect(ui->spExtrusionSpeed,SIGNAL(editingFinished()),this,SLOT(saveSettings()));
    connect(ui->spExtrusionSpeedAtStart,SIGNAL(editingFinished()),this,SLOT(saveSettings()));
	connect(ui->cbGenerateSupport,SIGNAL(toggled(bool)),this,SLOT(saveSettings()));
	connect(ui->cbMinimalSupport,SIGNAL(toggled(bool)),this,SLOT(saveSettings()));
    connect(ui->cbZmove,SIGNAL(toggled(bool)),this,SLOT(saveSettings()));
    connect(ui->spExtrudedWidth,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));
    connect(ui->spFillFactor,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));
    connect(ui->spZOffset,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));
    connect(ui->spExtrusionMultiplier,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));
	connect(ui->spRetractionOnMove,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));
    connect(ui->spSupportDensityRatio,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));
    connect(ui->spXmax,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));
    connect(ui->spYmax,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));
    connect(ui->spZmax,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));
    connect(ui->spXmax,SIGNAL(valueChanged(double)),this,SLOT(updateBuildRange()));
    connect(ui->spYmax,SIGNAL(valueChanged(double)),this,SLOT(updateBuildRange()));
    connect(ui->spZmax,SIGNAL(valueChanged(double)),this,SLOT(updateBuildRange()));
    connect(ui->bToggleHeat,SIGNAL(toggled(bool)),this,SLOT(toggleHeat(bool)));
    connect(ui->bGo,SIGNAL(pressed()),this,SLOT(handleGo()));
    connect(ui->bUpdate,SIGNAL(pressed()),this,SLOT(handleUpdateAxisPos()));
    connect(ui->bHomeAll,SIGNAL(pressed()),this,SLOT(handleHomeAll()));
	connect(ui->spMeshPosX,SIGNAL(valueChanged(double)),this,SLOT(updateTransform()));
	connect(ui->spMeshPosY,SIGNAL(valueChanged(double)),this,SLOT(updateTransform()));
	connect(ui->spMeshPosZ,SIGNAL(valueChanged(double)),this,SLOT(updateTransform()));
	connect(ui->spMeshRotX,SIGNAL(valueChanged(double)),this,SLOT(updateTransform()));
	connect(ui->spMeshRotY,SIGNAL(valueChanged(double)),this,SLOT(updateTransform()));
	connect(ui->spMeshRotZ,SIGNAL(valueChanged(double)),this,SLOT(updateTransform()));
	connect(ui->spMeshRotW,SIGNAL(valueChanged(double)),this,SLOT(updateTransform()));
	connect(ui->spMeshScale,SIGNAL(valueChanged(double)),this,SLOT(updateTransform()));
	connect(ui->bAutoCenter,SIGNAL(pressed()),this,SLOT(autoCenter()));
	connect(ui->bInvertNormals,SIGNAL(pressed()),Viewer::instance(),SLOT(invertNormals()));
    connect(ui->spShellOffset,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));
    connect(ui->spNbShells,SIGNAL(valueChanged(int)),this,SLOT(saveSettings()));
    connect(SerialConsole::instance(),SIGNAL(updateCoordinateX(double)),this,SLOT(setAxisPosX(double)));
    connect(SerialConsole::instance(),SIGNAL(updateCoordinateY(double)),this,SLOT(setAxisPosY(double)));
    connect(SerialConsole::instance(),SIGNAL(updateCoordinateZ(double)),this,SLOT(setAxisPosZ(double)));
    connect(ui->spXscale,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));
    connect(ui->spYscale,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));
    connect(ui->spZscale,SIGNAL(valueChanged(double)),this,SLOT(saveSettings()));

    connect(SerialConsole::instance(),SIGNAL(updateTemperature(double,double)),this,SLOT(addTemperatureSample(double,double)));
    connect(SerialConsole::instance(),SIGNAL(updateTemperature(double,double)),T_graph,SLOT(addTemperatureSample(double,double)));

    b_activate_extruder_notification = false;
}

PrintSettings::~PrintSettings()
{
    delete ui;
}

void PrintSettings::loadSettings()
{
    QSettings settings;
    settings.beginGroup("PrintSettings");
    ui->layer_height->setValue(settings.value("layer_height", ui->layer_height->value()).toDouble());
    ui->extruder_temperature->setValue(settings.value("extruder_temperature", ui->extruder_temperature->value()).toDouble());
    ui->bed_temperature->setValue(settings.value("bed_temperature", ui->bed_temperature->value()).toDouble());
    ui->spExtrusionSpeed->setValue(settings.value("extrusion_speed", ui->spExtrusionSpeed->value()).toDouble());
    ui->spExtrusionSpeedAtStart->setValue(settings.value("extrusion_speed_at_start", ui->spExtrusionSpeedAtStart->value()).toDouble());
	ui->cbGenerateSupport->setChecked(settings.value("generate_support", ui->cbGenerateSupport->isChecked()).toBool());
	ui->cbMinimalSupport->setChecked(settings.value("minimal_support", ui->cbMinimalSupport->isChecked()).toBool());
    ui->cbZmove->setChecked(settings.value("z_move", ui->cbZmove->isChecked()).toBool());
    ui->spExtrudedWidth->setValue(settings.value("extruded_width", ui->spExtrudedWidth->value()).toDouble());
    ui->spFillFactor->setValue(settings.value("fill_factor", ui->spFillFactor->value()).toDouble());
    ui->spZOffset->setValue(settings.value("z_offset", ui->spZOffset->value()).toDouble());
    ui->spExtrusionMultiplier->setValue(settings.value("extrusion_multiplier", ui->spExtrusionMultiplier->value()).toDouble());
	ui->spRetractionOnMove->setValue(settings.value("retraction_on_move", ui->spRetractionOnMove->value()).toDouble());
    ui->spSupportDensityRatio->setValue(settings.value("support_density_ratio", ui->spSupportDensityRatio->value()).toDouble());
    ui->spXmax->setValue(settings.value("x_max", ui->spXmax->value()).toDouble());
    ui->spYmax->setValue(settings.value("y_max", ui->spYmax->value()).toDouble());
    ui->spZmax->setValue(settings.value("z_max", ui->spZmax->value()).toDouble());
    ui->spShellOffset->setValue(settings.value("shell_offset", ui->spShellOffset->value()).toDouble());
    ui->spNbShells->setValue(settings.value("number_of_shells", ui->spNbShells->value()).toInt());
    ui->spXscale->setValue(settings.value("x_scale", ui->spXscale->value()).toDouble());
    ui->spYscale->setValue(settings.value("y_scale", ui->spYscale->value()).toDouble());
    ui->spZscale->setValue(settings.value("z_scale", ui->spZscale->value()).toDouble());
    settings.endGroup();
}

void PrintSettings::saveSettings()
{
    Viewer::instance()->updateGL();

    QSettings settings;
    settings.beginGroup("PrintSettings");
    settings.setValue("layer_height", ui->layer_height->value());
    settings.setValue("extruder_temperature", ui->extruder_temperature->value());
    settings.setValue("bed_temperature", ui->bed_temperature->value());
    settings.setValue("extrusion_speed", ui->spExtrusionSpeed->value());
    settings.setValue("extrusion_speed_at_start", ui->spExtrusionSpeedAtStart->value());
	settings.setValue("generate_support", ui->cbGenerateSupport->isChecked());
    settings.setValue("minimal_support", ui->cbMinimalSupport->isChecked());
    settings.setValue("z_move", ui->cbZmove->isChecked());
    settings.setValue("support_density_ratio", ui->spSupportDensityRatio->value());
    settings.setValue("extruded_width", ui->spExtrudedWidth->value());
    settings.setValue("fill_factor", ui->spFillFactor->value());
    settings.setValue("z_offset", ui->spZOffset->value());
	settings.setValue("extrusion_multiplier", ui->spExtrusionMultiplier->value());
	settings.setValue("retraction_on_move", ui->spRetractionOnMove->value());
    settings.setValue("x_max", ui->spXmax->value());
    settings.setValue("y_max", ui->spYmax->value());
    settings.setValue("z_max", ui->spZmax->value());
    settings.setValue("shell_offset", ui->spShellOffset->value());
    settings.setValue("number_of_shells", ui->spNbShells->value());
    settings.setValue("x_scale", ui->spXscale->value());
    settings.setValue("y_scale", ui->spYscale->value());
    settings.setValue("z_scale", ui->spZscale->value());
    settings.endGroup();
}

double PrintSettings::getLayerHeight() const
{
    return ui->layer_height->value();
}

double PrintSettings::getExtruderTargetTemperature() const
{
    return ui->extruder_temperature->value();
}

double PrintSettings::getBedTargetTemperature() const
{
    return ui->bed_temperature->value();
}

double PrintSettings::getExtrusionSpeed() const
{
    return ui->spExtrusionSpeed->value();
}

double PrintSettings::getSupportDensityRatio() const
{
    return ui->spSupportDensityRatio->value() / 100.0;
}

double PrintSettings::getExtrusionSpeedAtStart() const
{
    return ui->spExtrusionSpeedAtStart->value();
}

bool PrintSettings::isSupportEnabled() const
{
	return ui->cbGenerateSupport->isChecked();
}

bool PrintSettings::isMinimalSupportEnabled() const
{
	return ui->cbMinimalSupport->isChecked();
}

bool PrintSettings::isZmoveEnabled() const
{
    return ui->cbZmove->isChecked();
}

double PrintSettings::getFillFactor() const
{
    return ui->spFillFactor->value() / 100.0;
}

double PrintSettings::getExtrudedWidth() const
{
    return ui->spExtrudedWidth->value();
}

double PrintSettings::getZOffset() const
{
    return ui->spZOffset->value();
}

double PrintSettings::getExtrusionMultiplier() const
{
    return ui->spExtrusionMultiplier->value();
}

double PrintSettings::getRetractionOnMove() const
{
	return ui->spRetractionOnMove->value();
}

void PrintSettings::addTemperatureSample(double extruder_temperature, double bed_temperature)
{
    if (b_activate_extruder_notification
            && extruder_temperature >= ui->extruder_temperature->value() - 5
            && extruder_temperature < 500)
    {
        b_activate_extruder_notification = false;
        play_sound(":/sounds/complete.wav");
    }
}

void PrintSettings::toggleHeat(bool b)
{
    if (b)
    {
        b_activate_extruder_notification = true;
        ui->bToggleHeat->setText(tr("On"));
        ui->bToggleHeat->setChecked(true);
        QString msg;
        QTextStream out(&msg, QIODevice::WriteOnly);
        out << "M104 S" << getExtruderTargetTemperature();
        SerialConsole::instance()->sendMessage(msg.toLatin1());
        out << "M140 S" << getBedTargetTemperature();
        SerialConsole::instance()->sendMessage(msg.toLatin1());
    }
    else
    {
        b_activate_extruder_notification = false;
        ui->bToggleHeat->setText(tr("Off"));
        ui->bToggleHeat->setChecked(false);
        SerialConsole::instance()->sendMessage("M104 S0");
        SerialConsole::instance()->sendMessage("M140 S0");
    }
}

void PrintSettings::handleGo()
{
    QString msg;
    QTextStream out(&msg);
    out << "G0 X" << ui->spX->value() << " Y" << ui->spY->value() << " Z" << ui->spZ->value();
    SerialConsole::instance()->sendMessage(msg.toLatin1());
}

void PrintSettings::handleHomeAll()
{
    SerialConsole::instance()->sendMessage("G28");
}

void PrintSettings::handleUpdateAxisPos()
{
    SerialConsole::instance()->sendMessage("M114");
}

void PrintSettings::setAxisPos(double X, double Y, double Z)
{
    ui->spX->setValue(X);
    ui->spY->setValue(Y);
    ui->spZ->setValue(Z);
}

void PrintSettings::setAxisPosX(double X)
{
    ui->spX->setValue(X);
}

void PrintSettings::setAxisPosY(double Y)
{
    ui->spY->setValue(Y);
}

void PrintSettings::setAxisPosZ(double Z)
{
    ui->spZ->setValue(Z);
}

void PrintSettings::resetTransform()
{
	ui->spMeshPosX->setValue(0);
	ui->spMeshPosY->setValue(0);
	ui->spMeshPosZ->setValue(0);
	ui->spMeshRotX->setValue(0);
	ui->spMeshRotY->setValue(0);
	ui->spMeshRotZ->setValue(1);
	ui->spMeshRotW->setValue(0);
	ui->spMeshScale->setValue(1);

	Mesh *mesh = Viewer::instance()->getMesh();
	if (!mesh)
		return;
	const std::pair<Vec3, Vec3> &bbox = mesh->getBoundingBox();
	mesh_center = (bbox.first + bbox.second) / 2.0;
}

void PrintSettings::updateTransform()
{
	Mesh *mesh = Viewer::instance()->getMesh();
	if (!mesh)
		return;

	Vec3 t(ui->spMeshPosX->value(),
		   ui->spMeshPosY->value(),
		   ui->spMeshPosZ->value());
	Vec3 u(ui->spMeshRotX->value(),
		   ui->spMeshRotY->value(),
		   ui->spMeshRotZ->value());
	u.normalize();

	const Mat44 m = Mat44::translation(t + mesh_center)
					* Mat44::rotation(ui->spMeshRotW->value() * M_PI / 180.0, u)
					* Mat44::scale(Vec3(1,1,1) * ui->spMeshScale->value())
					* Mat44::translation(-mesh_center);

	Viewer::instance()->transformMesh(m);
}

void PrintSettings::autoCenter()
{
	Mesh *mesh = Viewer::instance()->getMesh();
	if (!mesh)
		return;

	updateTransform();

	const std::pair<Vec3, Vec3> &bbox = mesh->getBoundingBox();
	const std::pair<Vec3, Vec3> &buildrange = Viewer::instance()->getBuildRange();
	const Vec3 COBR = (buildrange.first + buildrange.second) / 2.0;
	const Vec3 COM = (bbox.second - bbox.first) / 2.0;
	Vec3 t(ui->spMeshPosX->value() - bbox.first.x() + COBR.x() - COM.x(),
		   ui->spMeshPosY->value() - bbox.first.y() + COBR.y() - COM.y(),
		   ui->spMeshPosZ->value() - bbox.first.z() + buildrange.first.z());

	ui->spMeshPosX->setValue(t.x());
	ui->spMeshPosY->setValue(t.y());
	ui->spMeshPosZ->setValue(t.z());
}

double PrintSettings::getXmax() const
{
    return ui->spXmax->value();
}

double PrintSettings::getYmax() const
{
    return ui->spYmax->value();
}

double PrintSettings::getZmax() const
{
    return ui->spZmax->value();
}

double PrintSettings::getXscale() const
{
    return ui->spXscale->value();
}

double PrintSettings::getYscale() const
{
    return ui->spYscale->value();
}

double PrintSettings::getZscale() const
{
    return ui->spZscale->value();
}

double PrintSettings::getShellOffset() const
{
    return ui->spShellOffset->value();
}

int PrintSettings::getNbShells() const
{
    return ui->spNbShells->value();
}

void PrintSettings::updateBuildRange() const
{
    const Vec3 lower(0,0,0);
    const Vec3 upper(getXmax(), getYmax(), getZmax());
    Viewer::instance()->setBuildRange(lower, upper);
}
