#include "mesh.h"
#include <unordered_map>
#include <unordered_set>
#include <QDebug>
#include <iostream>
#include <utility>
#include "kdtree.h"
#include "printsettings.h"
#include "generickdtree.h"

using namespace std;

namespace std
{
    template<typename T, typename U>
    struct hash<pair<T,U> >
    {
    public:
        inline size_t operator()(const pair<T,U> &v) const
        {
            return (h0(v.first) * 31U) ^ h1(v.second);
        }
    private:
        std::hash<T> h0;
        std::hash<U> h1;
    };
}

typedef pair<u_int32_t, u_int32_t> edge_t;

inline edge_t make_edge(const u_int32_t a, const u_int32_t b)
{
    return edge_t(min(a, b), max(a, b));
}

void Mesh::slice(const double z, const double h, vector<Vec2d> &lines, const int line_id) const
{
    lines.clear();

    unordered_map<edge_t, Vec2d> edges;
	unordered_map<edge_t, Vec2d> edge_normals;
	unordered_multimap<edge_t, edge_t> links;

	const double eps = 1e-6;//0.5 * h;
	const double extrusion_width = PrintSettings::instance()->getExtrudedWidth();
	const Vec2d no_point(numeric_limits<double>::infinity(), numeric_limits<double>::infinity());
    const Vec2d no_point_support(numeric_limits<double>::infinity(), -numeric_limits<double>::infinity());
    const double horizontal_cutoff_cos = 2.0 * extrusion_width / std::sqrt(4.0 * extrusion_width * extrusion_width + h * h);

    const double build_area_x_max = PrintSettings::instance()->getXmax();
    const double build_area_y_max = PrintSettings::instance()->getYmax();
    const double build_area_z_max = PrintSettings::instance()->getZmax();

    const double shell_offset = PrintSettings::instance()->getShellOffset();
    const int nb_shells = PrintSettings::instance()->getNbShells();

	// Build "vertical" external faces
    std::vector<edge_t> linked;
    kdtree->queryZ(z, [&](const Triangle &t)
    {
        const u_int32_t a = t.a;
        const u_int32_t b = t.b;
        const u_int32_t c = t.c;
        const double dz_a = vertex[a].z() - z;
        const double dz_b = vertex[b].z() - z;
        const double dz_c = vertex[c].z() - z;
        if (dz_a == 0.0 && dz_b == 0.0 && dz_c == 0.0)
            return;

		const size_t t_id = &t - ((const Triangle *)index.data());
		const Vec3d &f_n = face_normal[t_id];
		if (std::fabs(f_n.z()) >= horizontal_cutoff_cos)
			return;
		const Vec2d f_n2d(f_n.x(), f_n.y());

        linked.clear();
        bool b_a_set = false;
        bool b_b_set = false;
        bool b_c_set = false;
        if (std::fabs(dz_a) <= eps)
        {
            const edge_t e = make_edge(a, a);
            if (!edges.count(e))
			{
                edges.insert(make_pair(e, Vec2d(vertex[a].x(), vertex[a].y())));
				edge_normals.insert(make_pair(e, f_n2d));
			}
			else
				edge_normals[e] += f_n2d;
			linked.push_back(e);
            b_a_set = true;
        }
        if (std::fabs(dz_b) <= eps)
        {
            const edge_t e = make_edge(b, b);
            if (!edges.count(e))
			{
                edges.insert(make_pair(e, Vec2d(vertex[b].x(), vertex[b].y())));
				edge_normals.insert(make_pair(e, f_n2d));
			}
			else
				edge_normals[e] += f_n2d;
			linked.push_back(e);
            b_b_set = true;
        }
        if (std::fabs(dz_c) <= eps)
        {
            const edge_t e = make_edge(c, c);
            if (!edges.count(e))
			{
                edges.insert(make_pair(e, Vec2d(vertex[c].x(), vertex[c].y())));
				edge_normals.insert(make_pair(e, f_n2d));
			}
			else
				edge_normals[e] += f_n2d;
			linked.push_back(e);
            b_c_set = true;
        }
        if (dz_a * dz_b <= 0.0 && !b_a_set && !b_b_set)
        {
            const edge_t e = make_edge(a, b);
            if (!edges.count(e))
            {
                const Vec3 p = vertex[a] + (z - vertex[a].z()) / (vertex[b].z() - vertex[a].z()) * (vertex[b] - vertex[a]);
                edges.insert(make_pair(e, Vec2d(p.x(), p.y())));
				edge_normals.insert(make_pair(e, f_n2d));
			}
			else
				edge_normals[e] += f_n2d;
			linked.push_back(e);
        }
        if (dz_b * dz_c <= 0.0 && !b_b_set && !b_c_set)
        {
            const edge_t e = make_edge(b, c);
            if (!edges.count(e))
            {
                const Vec3 p = vertex[b] + (z - vertex[b].z()) / (vertex[c].z() - vertex[b].z()) * (vertex[c] - vertex[b]);
                edges.insert(make_pair(e, Vec2d(p.x(), p.y())));
				edge_normals.insert(make_pair(e, f_n2d));
			}
			else
				edge_normals[e] += f_n2d;
			linked.push_back(e);
        }
        if (dz_a * dz_c <= 0.0 && !b_a_set && !b_c_set)
        {
            const edge_t e = make_edge(a, c);
            if (!edges.count(e))
            {
                const Vec3 p = vertex[a] + (z - vertex[a].z()) / (vertex[c].z() - vertex[a].z()) * (vertex[c] - vertex[a]);
                edges.insert(make_pair(e, Vec2d(p.x(), p.y())));
				edge_normals.insert(make_pair(e, f_n2d));
			}
			else
				edge_normals[e] += f_n2d;
			linked.push_back(e);
        }
        for(size_t i = 0 ; i < linked.size() ; ++i)
            for(size_t j = i + 1 ; j < linked.size() ; ++j)
            {
                links.insert(make_pair(linked[i], linked[j]));
                links.insert(make_pair(linked[j], linked[i]));
            }
    });

    unordered_set<edge_t> work_queue;
    for(const auto &it : edges)
    {
        const edge_t &e = it.first;
        work_queue.insert(e);
    }
	for(auto &it : edge_normals)
		it.second.normalize();

	const edge_t not_an_edge = make_edge(numeric_limits<u_int32_t>::max(), numeric_limits<u_int32_t>::max());

	vector<edge_t> raw_lines;

    while(!work_queue.empty())
    {
        edge_t e = *work_queue.begin();

        if (!links.count(e))
        {
            work_queue.erase(e);
            continue;
        }
		if (!raw_lines.empty())
			raw_lines.push_back(not_an_edge);

		const size_t first_node_id = raw_lines.size();
        bool b_first_node = true;
        while(true)
        {
            if (!work_queue.count(e))
                break;
			const Vec2 c = edges[e];
			if (raw_lines.size() - first_node_id > 1)
            {
				const Vec2 a = edges[raw_lines[raw_lines.size() - 2]];
				const Vec2 b = edges[raw_lines[raw_lines.size() - 1]];
				Vec2 ab = b - a;
                Vec2 ac = c - a;
                ab.normalize();
                ac.normalize();
                const double det = std::fabs(ab.x() * ac.y() - ab.y() * ac.x());
                if (det < 1e-3)
                    raw_lines.back() = e;
				else
					raw_lines.push_back(e);
			}
            else
				raw_lines.push_back(e);
            if (!b_first_node)
                work_queue.erase(e);
            b_first_node = false;

            auto it = links.find(e);
            if (it == links.end())
                break;

            edge_t to = it->second;
            links.erase(it);
            auto range = links.equal_range(to);
            for(auto i = range.first ; i != range.second ; ++i)
                if (i->second == e)
                {
                    links.erase(i);
                    break;
                }
            e = to;
        }
		if (raw_lines.size() >= 4 + first_node_id)
		{
			const edge_t b = raw_lines[first_node_id];
			if (b == raw_lines.back())
			{
				const Vec2 _b = edges[b];
				const edge_t c = raw_lines[first_node_id + 1];
				const edge_t a = raw_lines[raw_lines.size() - 2];
				const Vec2 _c = edges[c];
				const Vec2 _a = edges[a];
				Vec2 ab = _b - _a;
				Vec2 ac = _c - _a;
				if (ab.sq() > 0.0 && ac.sq() > 0.0)
				{
					const double _ab = ab.length();
					ab.normalize();
					ac.normalize();
					const double det = std::fabs(ab.x() * ac.y() - ab.y() * ac.x());
					if (det < 1e-2 || det * _ab < 0.1)
					{
						raw_lines[first_node_id] = a;
						raw_lines.pop_back();
					}
				}
			}
		}
	}

	// Add an offset to compensate for the filament width
	if (!raw_lines.empty())
	{
		lines.reserve(raw_lines.size() * 2 + 1);
        for(int shell_id = 0 ; shell_id < nb_shells ; ++shell_id)
        {
            const double total_offset = shell_offset + extrusion_width * (shell_id + 0.5);
            for(auto &e : raw_lines)
            {
                if (e == not_an_edge)
                {
                    lines.push_back(no_point);
                    continue;
                }
                const Vec2d n = edge_normals[e];
                const Vec2d p = edges[e];
                lines.push_back(p - total_offset * n);
            }
            if (shell_id + 1 < nb_shells)
                lines.push_back(no_point);
        }
	}

	// Build top and bottom polygons
	const double z_min = (z - h) * (1.0 - std::numeric_limits<double>::epsilon());
	const double z_max = (z + h) * (1.0 + std::numeric_limits<double>::epsilon());
	std::vector<Vec2> triangles_to_fill;
	kdtree->queryZ(z_min, z_max, [&](const Triangle &t)
	{
		const size_t t_id = &t - ((const Triangle *)index.data());
		const Vec3d &f_n = face_normal[t_id];
		if (std::fabs(f_n.z()) < horizontal_cutoff_cos)
			return;
		Vec3 a = vertex[t.a];
		Vec3 b = vertex[t.b];
		Vec3 c = vertex[t.c];
		if (a.z() > b.z())	std::swap(a, b);
		if (b.z() > c.z())	std::swap(b, c);
		if (a.z() > b.z())	std::swap(a, b);
		if (f_n.z() < 0.0 && a.z() > z + h * 0.5)
            return;
		if (f_n.z() > 0.0 && c.z() < z - 0.5 * h)
            return;

		if (a.z() >= z_min)
		{
			if (c.z() <= z_max)
			{
				triangles_to_fill.push_back(Vec2(a.x(), a.y()));
				triangles_to_fill.push_back(Vec2(b.x(), b.y()));
				triangles_to_fill.push_back(Vec2(c.x(), c.y()));
			}
			else if (b.z() <= z_max)
			{
				const Vec3 ac = a + (z_max - a.z()) / (c.z() - a.z()) * (c - a);
				const Vec3 bc = c.z() - b.z() > 1e-6 ? b + (z_max - b.z()) / (c.z() - b.z()) * (c - b) : c;
				triangles_to_fill.push_back(Vec2(a.x(), a.y()));
				triangles_to_fill.push_back(Vec2(b.x(), b.y()));
				triangles_to_fill.push_back(Vec2(ac.x(), ac.y()));
				triangles_to_fill.push_back(Vec2(ac.x(), ac.y()));
				triangles_to_fill.push_back(Vec2(b.x(), b.y()));
				triangles_to_fill.push_back(Vec2(bc.x(), bc.y()));
			}
			else
			{
				const Vec3 ac = a + (z_max - a.z()) / (c.z() - a.z()) * (c - a);
				const Vec3 ab = a + (z_max - a.z()) / (b.z() - a.z()) * (b - a);
				triangles_to_fill.push_back(Vec2(a.x(), a.y()));
				triangles_to_fill.push_back(Vec2(ab.x(), ab.y()));
				triangles_to_fill.push_back(Vec2(ac.x(), ac.y()));
			}
		}
		else
		{
			if (b.z() >= z_min)
			{
				if (c.z() <= z_max)
				{
					const Vec3 ac = c.z() - a.z() > 1e-6 ? a + (z_min - a.z()) / (c.z() - a.z()) * (c - a) : c;
					const Vec3 ab = b.z() - a.z() > 1e-6 ? a + (z_min - a.z()) / (b.z() - a.z()) * (b - a) : b;
					triangles_to_fill.push_back(Vec2(ab.x(), ab.y()));
					triangles_to_fill.push_back(Vec2(b.x(), b.y()));
					triangles_to_fill.push_back(Vec2(c.x(), c.y()));

					triangles_to_fill.push_back(Vec2(ab.x(), ab.y()));
					triangles_to_fill.push_back(Vec2(c.x(), c.y()));
					triangles_to_fill.push_back(Vec2(ac.x(), ac.y()));
				}
				else if (b.z() <= z_max)
				{
					const Vec3 ac0 = a + (z_min - a.z()) / (c.z() - a.z()) * (c - a);
					const Vec3 ac1 = a + (z_max - a.z()) / (c.z() - a.z()) * (c - a);
					const Vec3 ab = b.z() - a.z() > 1e-6 ? a + (z_min - a.z()) / (b.z() - a.z()) * (b - a) : b;
					const Vec3 bc = c.z() - b.z() > 1e-6 ? b + (z_max - b.z()) / (c.z() - b.z()) * (c - b) : b;

					triangles_to_fill.push_back(Vec2(ac0.x(), ac0.y()));
					triangles_to_fill.push_back(Vec2(ab.x(), ab.y()));
					triangles_to_fill.push_back(Vec2(bc.x(), bc.y()));

					triangles_to_fill.push_back(Vec2(ac0.x(), ac0.y()));
					triangles_to_fill.push_back(Vec2(ac1.x(), ac1.y()));
					triangles_to_fill.push_back(Vec2(bc.x(), bc.y()));
				}
				else
				{
					const Vec3 ac0 = a + (z_min - a.z()) / (c.z() - a.z()) * (c - a);
					const Vec3 ac1 = a + (z_max - a.z()) / (c.z() - a.z()) * (c - a);
					const Vec3 ab0 = a + (z_min - a.z()) / (b.z() - a.z()) * (b - a);
					const Vec3 ab1 = a + (z_max - a.z()) / (b.z() - a.z()) * (b - a);

					triangles_to_fill.push_back(Vec2(ac0.x(), ac0.y()));
					triangles_to_fill.push_back(Vec2(ac1.x(), ac1.y()));
					triangles_to_fill.push_back(Vec2(ab0.x(), ab0.y()));

					triangles_to_fill.push_back(Vec2(ab1.x(), ab1.y()));
					triangles_to_fill.push_back(Vec2(ac1.x(), ac1.y()));
					triangles_to_fill.push_back(Vec2(ab0.x(), ab0.y()));
				}
			}
			else
			{
				if (c.z() <= z_max)
				{
					const Vec3 ac = c.z() - a.z() > 1e-6 ? a + (z_min - a.z()) / (c.z() - a.z()) * (c - a) : a;
					const Vec3 bc = c.z() - b.z() > 1e-6 ? b + (z_min - b.z()) / (c.z() - b.z()) * (c - b) : b;
					triangles_to_fill.push_back(Vec2(ac.x(), ac.y()));
					triangles_to_fill.push_back(Vec2(bc.x(), bc.y()));
					triangles_to_fill.push_back(Vec2(c.x(), c.y()));
				}
				else
				{
					const Vec3 ac0 = a + (z_min - a.z()) / (c.z() - a.z()) * (c - a);
					const Vec3 ac1 = a + (z_max - a.z()) / (c.z() - a.z()) * (c - a);
					const Vec3 bc0 = b + (z_min - b.z()) / (c.z() - b.z()) * (c - b);
					const Vec3 bc1 = b + (z_max - b.z()) / (c.z() - b.z()) * (c - b);

					triangles_to_fill.push_back(Vec2(ac0.x(), ac0.y()));
					triangles_to_fill.push_back(Vec2(ac1.x(), ac1.y()));
					triangles_to_fill.push_back(Vec2(bc0.x(), bc0.y()));

					triangles_to_fill.push_back(Vec2(bc1.x(), bc1.y()));
					triangles_to_fill.push_back(Vec2(ac1.x(), ac1.y()));
					triangles_to_fill.push_back(Vec2(bc0.x(), bc0.y()));
				}
			}
		}
	});
	if (!triangles_to_fill.empty())
	{
		std::unordered_map<int, std::vector<std::pair<double, bool> > > segments;
		Mat22 R(Mat22::rotation(-line_id * M_PI / 3.0));
		R(1,0) /= extrusion_width;
		R(1,1) /= extrusion_width;
		for(size_t i = 0 ; i < triangles_to_fill.size() ; i += 3)
		{
			Vec2d a = R * triangles_to_fill[i];
			Vec2d b = R * triangles_to_fill[i + 1];
			Vec2d c = R * triangles_to_fill[i + 2];
			if (a.y() > b.y())	std::swap(a, b);
			if (b.y() > c.y())	std::swap(b, c);
			if (a.y() > b.y())	std::swap(a, b);

			for(int y = std::ceil(a.y()) ; y < b.y() ; ++y)
			{
				double x0 = a.x() + (b.x() - a.x()) * (y - a.y()) / (b.y() - a.y());
				double x1 = a.x() + (c.x() - a.x()) * (y - a.y()) / (c.y() - a.y());
				if (x0 > x1)	std::swap(x0, x1);
				auto &v = segments[y];
				v.push_back(make_pair(x0, true));
				v.push_back(make_pair(x1, false));
			}
			for(int y = std::ceil(b.y()) ; y <= c.y() ; ++y)
			{
				double x0 = b.x() + (c.x() - b.x()) * (y - b.y()) / (c.y() - b.y());
				double x1 = a.x() + (c.x() - a.x()) * (y - a.y()) / (c.y() - a.y());
				if (x0 > x1)	std::swap(x0, x1);
				auto &v = segments[y];
				v.push_back(make_pair(x0, true));
				v.push_back(make_pair(x1, false));
			}
		}
		R = Mat22::rotation(line_id * M_PI / 3.0);
		R(0,1) *= extrusion_width;
		R(1,1) *= extrusion_width;

		for(auto &s_y : segments)
		{
			const int y = s_y.first;
			std::sort(s_y.second.begin(), s_y.second.end(),
					  [](const std::pair<double, bool> &a, const std::pair<double, bool> &b)
			{
				return a.first < b.first;
			});
            int n = 0;
			Vec2 start;
			double prev_l(0);
			for(size_t i = 0 ; i < s_y.second.size() ; ++i)
			{
				const auto &it = s_y.second[i];
				if (it.second)
				{
					if (n == 0)
					{
                        prev_l = it.first + (nb_shells - 0.5) * extrusion_width;
						start = R * Vec2(prev_l, y);
					}
					++n;
				}
				else
				{
					--n;
					if (n == 0)
					{
						if (i + 1 < s_y.second.size() && s_y.second[i + 1].first == it.first)
						{
							++n;
							++i;
							continue;
						}
                        const double next_l = it.first - (nb_shells - 0.5) * extrusion_width;
						if (next_l <= prev_l)
							continue;
						if (!lines.empty())
							lines.push_back(no_point);
						lines.push_back(start);
						lines.push_back(R * Vec2(next_l, y));
					}
				}
			}
		}
	}

	// Build the interior
    const double fill_factor = PrintSettings::instance()->getFillFactor();
    if (fill_factor > 0.0)
    {
        const double step = extrusion_width / fill_factor;
		const Vec3 dir(std::cos(line_id * M_PI / 3.0), std::sin(line_id * M_PI / 3.0), 0.0);
		const Vec2 dir2(std::cos(line_id * M_PI / 3.0), std::sin(line_id * M_PI / 3.0));
		const Vec3 side(std::sin(line_id * M_PI / 3.0), -std::cos(line_id * M_PI / 3.0), 0.0);
		const Vec2 side2(std::sin(line_id * M_PI / 3.0), -std::cos(line_id * M_PI / 3.0));
		double begin = std::numeric_limits<double>::infinity();
        double end = -std::numeric_limits<double>::infinity();
        double front = std::numeric_limits<double>::infinity();
        double back = -std::numeric_limits<double>::infinity();
        for(const auto &it : edges)
        {
            const double v = it.second | side2;
            const double v2 = it.second | dir2;
            begin = std::min(v, begin);
            end = std::max(v, end);
            front = std::min(v2, front);
            back = std::max(v2, back);
        }
        front -= 1.0;
        back += 1.0;

        std::unordered_set<const Triangle*> avoid;
        bool b_start_from_front = true;
        for(double l = begin + 0.5 * step ; l <= end ; l += step)
        {
            Vec3 pos, rdir;
            if (b_start_from_front)
            {
                pos = front * dir + l * side + Vec3(0,0,z);
                rdir = dir;
            }
            else
            {
                pos = back * dir + l * side + Vec3(0,0,z);
                rdir = -dir;
            }
            Vec3 hit_point = pos;
			bool b_fill = false;
            avoid.clear();
            const Triangle *t_hit;
            bool b_hit = false;
            while ((t_hit = kdtree->queryRay(pos, rdir, hit_point, avoid)))
            {
                b_hit = true;
                avoid.insert(t_hit);
                if (b_fill && ((hit_point - pos) | rdir) > extrusion_width * 2.0 * nb_shells)
                {
					lines.push_back(no_point);
                    pos = pos + nb_shells * extrusion_width * rdir;
                    lines.push_back(Vec2d(pos.x(), pos.y()));
                    pos = hit_point - nb_shells * extrusion_width * rdir;
                    lines.push_back(Vec2d(pos.x(), pos.y()));
                }
				const Vec3 &n = face_normal[t_hit - (const Triangle*)index.data()];
				if ((n | rdir) < 0.0f)
					b_fill = true;
				else if ((n | rdir) > 0.0f)
					b_fill = false;
				pos = hit_point;
            }
            if (b_hit)
                b_start_from_front ^= true;
        }
    }

	// Build the support
	if (PrintSettings::instance()->isSupportEnabled())
	{
		const double step = 2.0;
		const Vec3 dir(1.0, 0.0, 0.0);
		const Vec3 side(0.0, -1.0, 0.0);
		double begin = -bbox.second.y();
		double end = -bbox.first.y();
		double front = bbox.first.x();
		double back = bbox.second.x();
		front -= 1.0;
		back += 1.0;

		std::unordered_set<const Triangle*> avoid;
		bool b_start_from_front = true;
		for(double l = begin + 0.5 * step ; l <= end ; l += step)
		{
			Vec3 pos, rdir;
			if (b_start_from_front)
			{
				pos = front * dir + l * side + Vec3(0,0,z);
				rdir = dir;
			}
			else
			{
				pos = back * dir + l * side + Vec3(0,0,z);
				rdir = -dir;
			}
			Vec3 hit_point = pos;
			bool b_inside = false;
			avoid.clear();
			const Triangle *t_hit = hull_kdtree->queryRay(pos, rdir, hit_point, avoid);
			if (!t_hit)
				continue;
			bool b_hit = false;
			bool b_last_step_done = false;
			while ((t_hit = kdtree->queryRay(pos, rdir, hit_point, avoid)) || !b_hit || !b_last_step_done)
			{
				if (t_hit)
				{
					b_hit = true;
					avoid.insert(t_hit);
				}
				else
				{
					if (!b_hit)
					{
						hit_point = pos + (back - front) * rdir;
						b_last_step_done = true;
					}
					else if (!b_last_step_done)
					{
						b_last_step_done = true;
						hit_point = pos + (back - front) * rdir;
					}
				}
				if (!b_inside
                        && ((hit_point - pos) | rdir) > extrusion_width * 2.0)
				{
                    const double l = ((hit_point - pos) | rdir) - extrusion_width * 2.0;
                    Vec3 start = pos + 1.0 * extrusion_width * rdir;
					bool b_last_was_ok = false;
					Vec3 last_start = start;
					Vec3 last_P = start;
					for(double k = 0.0 ; k <= l ; k += extrusion_width * 4.0)
					{
						const Vec3 P = start + k * rdir;
						Vec3 hit_z;
						const Triangle *t_hit_z = base_kdtree->queryRay(P, Vec3(0,0,1), hit_z);
                        if (t_hit_z && hit_z.z() - pos.z() > h && hit_z.z() <= build_area_z_max)
						{
							if (!b_last_was_ok)
							{
								last_start = P;
							}
							b_last_was_ok = true;
						}
						else
						{
							if (b_last_was_ok)
							{
                                lines.push_back(no_point_support);
								lines.push_back(Vec2d(last_start.x(), last_start.y()));
								lines.push_back(Vec2d(last_P.x(), last_P.y()));
							}
							b_last_was_ok = false;
						}
						last_P = P;
					}
					if (b_last_was_ok && (last_P - last_start).sq() > 0.0)
					{
                        lines.push_back(no_point_support);
						lines.push_back(Vec2d(last_start.x(), last_start.y()));
						lines.push_back(Vec2d(last_P.x(), last_P.y()));
					}
				}
				b_inside ^= true;
				pos = hit_point;
				if (!b_hit)
					break;
			}
			if (b_hit)
				b_start_from_front ^= true;
		}
	}
}

void Mesh::buildSupportHull()
{
	const double extrusion_width = PrintSettings::instance()->getExtrudedWidth();
	const double h = PrintSettings::instance()->getLayerHeight();
	const double cutoff_cos = extrusion_width * 2.0 / std::sqrt(4.0 * extrusion_width * extrusion_width + h * h);

	base_vertex.clear();
	base_index.clear();
	if (PrintSettings::instance()->isMinimalSupportEnabled())
	{
		std::vector<bool> lower_points;
		lower_points.resize(vertex.size(), true);
		for(size_t i = 0 ; i < index.size() ; i += 3)
		{
			const size_t _a = index[i];
			const size_t _b = index[i + 1];
			const size_t _c = index[i + 2];
			const Vec3 &a = vertex[_a];
			const Vec3 &b = vertex[_b];
			const Vec3 &c = vertex[_c];
			if (a.z() > b.z() || a.z() > c.z())
				lower_points[_a] = false;
			if (b.z() > a.z() || b.z() > c.z())
				lower_points[_b] = false;
			if (c.z() > a.z() || c.z() > b.z())
				lower_points[_c] = false;
		}
		std::unordered_map<size_t, size_t> mapping;
		for(size_t i = 0 ; i < index.size() ; i += 3)
		{
			const Vec3 &f_n = face_normal[i / 3];
			if (f_n.z() > 0.0)
				continue;
			const size_t _a = index[i];
			const size_t _b = index[i + 1];
			const size_t _c = index[i + 2];
			if (lower_points[_a] || lower_points[_b] || lower_points[_c] || -f_n.z() >= cutoff_cos)
			{
				const auto it_a = mapping.find(_a);
				if (it_a != mapping.end())
					base_index.push_back(it_a->second);
				else
				{
					base_index.push_back(mapping[_a] = base_vertex.size());
					base_vertex.push_back(vertex[_a]);
				}
				const auto it_b = mapping.find(_b);
				if (it_b != mapping.end())
					base_index.push_back(it_b->second);
				else
				{
					base_index.push_back(mapping[_b] = base_vertex.size());
					base_vertex.push_back(vertex[_b]);
				}
				const auto it_c = mapping.find(_c);
				if (it_c != mapping.end())
					base_index.push_back(it_c->second);
				else
				{
					base_index.push_back(mapping[_c] = base_vertex.size());
					base_vertex.push_back(vertex[_c]);
				}
			}
		}
	}
	else
	{
		std::unordered_map<size_t, size_t> mapping;
		for(size_t i = 0 ; i < index.size() ; i += 3)
		{
			const Vec3 &f_n = face_normal[i / 3];
			if (f_n.z() > 0.0)
				continue;
			const size_t _a = index[i];
			const size_t _b = index[i + 1];
			const size_t _c = index[i + 2];
			const auto it_a = mapping.find(_a);
			if (it_a != mapping.end())
				base_index.push_back(it_a->second);
			else
			{
				base_index.push_back(mapping[_a] = base_vertex.size());
				base_vertex.push_back(vertex[_a]);
			}
			const auto it_b = mapping.find(_b);
			if (it_b != mapping.end())
				base_index.push_back(it_b->second);
			else
			{
				base_index.push_back(mapping[_b] = base_vertex.size());
				base_vertex.push_back(vertex[_b]);
			}
			const auto it_c = mapping.find(_c);
			if (it_c != mapping.end())
				base_index.push_back(it_c->second);
			else
			{
				base_index.push_back(mapping[_c] = base_vertex.size());
				base_vertex.push_back(vertex[_c]);
			}
		}
	}

	hull_vertex.clear();
	hull_index.clear();

	unordered_multimap<edge_t, int> edge2tri;
	unordered_set<edge_t> edges;

	for(size_t i = 0 ; i < base_index.size() ; i += 3)
	{
		const edge_t ab = make_edge(base_index[i], base_index[i + 1]);
		const edge_t bc = make_edge(base_index[i + 1], base_index[i + 2]);
		const edge_t ac = make_edge(base_index[i], base_index[i + 2]);
		edge2tri.insert(make_pair(ab, i));
		edge2tri.insert(make_pair(bc, i));
		edge2tri.insert(make_pair(ac, i));
		edges.insert(ab);
		edges.insert(bc);
		edges.insert(ac);
	}

	for(const edge_t &e : edges)
	{
		const auto &range = edge2tri.equal_range(e);
		bool b_part_of_hull = false;
		const Vec3 &A = base_vertex[e.first];
		const Vec3 &B = base_vertex[e.second];
		auto et = range.first;
		int t0 = et->second;
		for(++et ; et != range.second && !b_part_of_hull ; ++et)
		{
			const int t1 = et->second;
			int c = -1, d = -1;
			for(int i = 0 ; i < 3 ; ++i)
			{
				if (base_index[t0 + i] != e.first && base_index[t0 + i] != e.second)
					c = base_index[t0 + i];
				if (base_index[t1 + i] != e.first && base_index[t1 + i] != e.second)
					d = base_index[t1 + i];
			}
			t0 = t1;
			if (c == -1 || d == -1)
				continue;
			const Vec3 &C = base_vertex[c];
			const Vec3 &D = base_vertex[d];
			b_part_of_hull = (((B - A) ^ (C - A)) | ((B - A) ^ (D - A))) <= 0.0;
		}
		if (!b_part_of_hull)
			continue;

		const int a = hull_vertex.size();
		hull_vertex.push_back(A);
		const int b = hull_vertex.size();
		hull_vertex.push_back(B);

		Vec3 C(A), D(B);
		C.z() = bbox.first.z() - 1.0e3;
		D.z() = bbox.first.z() - 1.0e3;
		const int c = hull_vertex.size();
		hull_vertex.push_back(C);
		const int d = hull_vertex.size();
		hull_vertex.push_back(D);

		hull_index.push_back(a);
		hull_index.push_back(b);
		hull_index.push_back(c);
		hull_index.push_back(b);
		hull_index.push_back(c);
		hull_index.push_back(d);
	}

	if (hull_kdtree)
		delete hull_kdtree;
	if (base_kdtree)
		delete base_kdtree;
#pragma omp parallel
	{
#pragma omp single
		hull_kdtree = new KdTree(hull_vertex.data(), (Triangle*)hull_index.data(), (Triangle*)hull_index.data() + hull_index.size() / 3);
#pragma omp single
		base_kdtree = new KdTree(base_vertex.data(), (Triangle*)base_index.data(), (Triangle*)base_index.data() + base_index.size() / 3);
	}
}

void Mesh::computeFaceNormals()
{
	face_normal.clear();
	face_normal.resize(index.size() / 3);

#pragma omp parallel for
	for(size_t i = 0 ; i < index.size() ; i += 3)
	{
		const size_t a = index[i];
		const size_t b = index[i + 1];
		const size_t c = index[i + 2];
		const Vec3 &A = vertex[a];
		const Vec3 &B = vertex[b];
		const Vec3 &C = vertex[c];
		Vec3 n = (B - A) ^ (C - A);
		n.normalize();
		face_normal[i / 3] = n;
	}
}

Vec2d Mesh::optimizeSlice(const Vec2d &start, std::vector<Vec2d> &lines)
{
    if (lines.empty())
		return start;
	const Vec2d no_point(numeric_limits<double>::infinity(), numeric_limits<double>::infinity());
    const Vec2d no_point_support(numeric_limits<double>::infinity(), -numeric_limits<double>::infinity());
    // optimize extruder path
	std::vector<std::pair<size_t, std::pair<Vec2d, Vec2d> > > segments;
	bool b_start_new_line = true;
	size_t line_start = -1;
	for(size_t i = 0 ; i < lines.size() ; ++i)
	{
		if (lines[i].x() == numeric_limits<double>::infinity())
		{
			if (line_start < i)
            {
                if (!std::isnan(lines[line_start].x()) && !std::isnan(lines[line_start].y()) && !std::isnan(lines[i-1].x()) && !std::isnan(lines[i-1].y()))
                    segments.push_back(std::make_pair(line_start, std::make_pair(lines[line_start], lines[i-1])));
            }
			line_start = -1;
			b_start_new_line = true;
			continue;
		}
		else if (b_start_new_line)
		{
			line_start = i;
			b_start_new_line = false;
		}
	}
    if (line_start < lines.size() - 1)
        segments.push_back(std::make_pair(line_start, std::make_pair(lines[line_start], lines[lines.size()])));

	std::vector<std::pair<Vec2d, size_t> > pts;
	pts.reserve(segments.size() * 2);
	for(size_t i = 0 ; i < segments.size() ; ++i)
	{
		const auto &s = segments[i];
		pts.push_back(make_pair(s.second.first, i));
		pts.push_back(make_pair(s.second.second, i));
	}
	GenericKdTree<std::pair<Vec2d, size_t>, N2> pts_tree(pts);

	std::vector<std::pair<size_t, size_t> > mapping;
    mapping.resize(segments.size(), make_pair<size_t>(-1, -1));
	for(size_t i = 0 ; i < pts.size() ; ++i)
		if (mapping[pts[i].second].first == size_t(-1))
			mapping[pts[i].second].first = i;
		else
			mapping[pts[i].second].second = i;

	std::vector<Vec2d> out;
	out.reserve(lines.size());

	Vec2d last = lines.front();
	if (start.x() != std::numeric_limits<double>::infinity())
		last = start;
	while(true)
	{
		std::pair<Vec2d, size_t> *next = pts_tree.nearest_if(make_pair(last, size_t(-1)), [](const std::pair<Vec2d, size_t> &v)	{	return v.second != size_t(-1);	});
		if (!next)
			break;
		const size_t segment_id = next->second;
		pts[mapping[segment_id].first].second = -1;
		pts[mapping[segment_id].second].second = -1;

        const size_t start = segments[segment_id].first;
        if (start > 0 && lines[start-1].y() == -numeric_limits<double>::infinity())
            out.push_back(no_point_support);
        else
            if (!out.empty())
                out.push_back(no_point);

		if (next->first == lines[start])	// Forward
		{
			for(size_t i = start ; i < lines.size() && lines[i].x() != numeric_limits<double>::infinity() ; ++i)
				out.push_back(lines[i]);
		}
		else		// Backward
		{
			int i;
			for(i = start ; i < lines.size() && lines[i].x() != numeric_limits<double>::infinity() ; ++i);
			for(--i ; i >= int(start) ; --i)
				out.push_back(lines[i]);
		}
		last = out.back();
	}
    lines.swap(out);
	return last;
}
