#include "utils.h"
#include "mesh.h"
#include <QAudioDeviceInfo>
#include <QAudioOutput>
#include <QAudioFormat>
#include <QFile>
#include <QDebug>
#include <QTimer>

bool hit_sphere(const Vec3 &pos, const Vec3 &dir, const Vec3 &center, const double radius, Vec3 &hit_point)
{
	const Vec3 &OC = center - pos;
    const double l = OC | dir;
	const Vec3 &p = pos + l * dir;
    const double d2 = (p - center).sq();
    const double r2 = radius * radius;
	if (d2 > r2)
		return false;
	hit_point = pos + (l - std::sqrt(r2 - d2)) * dir;
	return true;
}

struct WAV_header
{
    quint32 FileTypeBlocID;     // RIFF
    quint32 FileSize;           // file size - 8 bytes
    quint32 FileFormatID;       // WAVE
};

struct WAV_audio_format
{
   quint32 FormatBlocID;        // Identifiant «fmt »  (0x66,0x6D, 0x74,0x20)
   quint32 BlocSize;            // Nombre d'octets du bloc - 8  (0x10)

   quint16 AudioFormat;         // Format du stockage dans le fichier (1: PCM, ...)
   quint16 NbrCanaux;           // Nombre de canaux (de 1 à 6, cf. ci-dessous)
   quint32 Frequence;           // Fréquence d'échantillonnage (en hertz) [Valeurs standardisées : 11025, 22050, 44100 et éventuellement 48000 et 96000]
   quint32 BytePerSec;          // Nombre d'octets à lire par seconde (i.e., Frequence * BytePerBloc).
   quint16 BytePerBloc;         // Nombre d'octets par bloc d'échantillonnage (i.e., tous canaux confondus : NbrCanaux * BitsPerSample/8).
   quint16 BitsPerSample;       // Nombre de bits utilisés pour le codage de chaque échantillon (8, 16, 24)
};

struct WAV_data_bloc_header
{
   quint32 DataBlocID;      // Constante «data»  (0x64,0x61,0x74,0x61)
   quint32 DataSize;        // Nombre d'octets des données (i.e. "Data[]", i.e. taille_du_fichier - taille_de_l'entête  (qui fait 44 octets normalement).
//   DATAS[] : [Octets du Sample 1 du Canal 1] [Octets du Sample 1 du Canal 2] [Octets du Sample 2 du Canal 1] [Octets du Sample 2 du Canal 2]
};

void play_sound(const QString &filename)
{
    QAudioDeviceInfo devID = QAudioDeviceInfo::defaultOutputDevice();

    QFile *file = new QFile(filename);
    file->open(QIODevice::ReadOnly);

    WAV_header header;
    WAV_audio_format format;
    WAV_data_bloc_header data_header;
    file->read((char*)&header, sizeof(header));
    if (header.FileTypeBlocID != 0x46464952U)
    {
        delete file;
        return;
    }
    if (header.FileFormatID != 0x45564157U)
    {
        delete file;
        return;
    }
    file->read((char*)&format, sizeof(format));
    if (format.FormatBlocID != 0x20746D66U)
    {
        delete file;
        return;
    }
    if (format.AudioFormat != 1)
    {
        delete file;
        return;
    }
    file->read((char*)&data_header, sizeof(data_header));

    if (data_header.DataBlocID != 0x61746164U)
        return;

    QAudioFormat fmt;
    fmt.setChannelCount(format.NbrCanaux);
#if QT_VERSION >= QT_VERSION_CHECK(5,0,0)
    fmt.setSampleRate(format.Frequence);
#else
    fmt.setFrequency(format.Frequence);
#endif
    fmt.setSampleSize(format.BitsPerSample);
    fmt.setSampleType(QAudioFormat::SignedInt);
    fmt.setByteOrder(QAudioFormat::LittleEndian);
    fmt.setCodec("audio/pcm");

    QAudioOutput *dev = new QAudioOutput(devID, fmt);
    dev->start(file);

    const double duration = double(data_header.DataSize) / format.BytePerBloc / format.Frequence;
    QTimer *timer = new QTimer;
    timer->setInterval(duration * 1000.0 + 1.0);
    timer->setSingleShot(true);
    QObject::connect(timer,SIGNAL(timeout()),dev,SLOT(deleteLater()));
    QObject::connect(timer,SIGNAL(timeout()),file,SLOT(deleteLater()));
    QObject::connect(timer,SIGNAL(timeout()),timer,SLOT(deleteLater()));
}
